function [Z,D,W] = sainv (A, p, tol1, tol2)
% SAINV: compute (stabilized) factored approximate inverse preconditioner
% Usage for symmetric matrices:
%    [Z,D] = sainv (A, p, tol1, tol2);
% Usage for unsymmetric matrices:
%    [Z,D,W] = sainv (A, p, tol1, tol2);
% Here A is the sparse square matrix you want to precondition, and the
% the approximate inverse is given by Z*D*Z' (symmetric) or Z*D*W' (otherwise).
% In particular, W'*A*Z*D is guaranteed to have unit diagonal if no breakdown
% occurs (which is impossible if A is symmetric positive definite).
% W and Z are logically upper triangular and D is diagonal.
%
% p is optional, and determines the ordering used. If not given, the default
% symmmd ordering is used (you may want to change this to a high quality
% fill-reducing ordering based on nested dissection or a good implementation
% of minimum degree if you can). If "0" is given, the original ordering of
% A is accepted, so W and Z will truly be uper triangular. If a permutation
% vector is given, that is used instead.
%
% tol1 and tol2 are optional, and control dropping in the algorithm. tol1
% is used to drop from the rank-1 updates (to the dynamic data structure) and
% tol2 is then used to filter out small entries from there. If not given,
% they default to tol1==1e-9 and tol2==0.1. For a given tol2, you can get
% slightly higher speed and better memory usage by selecting tol1 as large as
% or comparable to tol2, but it may have a bad effect on preconditioner
% quality.

if nargin < 1
   error ('need to specify a matrix');
end
A = sparse(A);   % make sure it's sparse
if nargin < 2
   if size(A,1) ~= size(A,2)
      error ('matrix must be square');
   end
   p = symmmd(A);   % get default ordering
   A = A(p,p);
   q(p) = 1:length(A);
else
   if isempty(p)
      q = 1:length(A);
   else
      A = A(p,p);
      q(p) = 1:length(A);
   end
end
if nargin < 3
   tol1 = 1e-9;
end
if nargin < 4
   tol2 = 0.1;
end

if nargout == 2
   [Z,D] = sainvmex(tol1,tol2,A);
elseif nargout == 3
   [Z,D,W] = sainvmex(tol1,tol2,A,A');
else
   error ('for symmetric A specify two output arguments, otherwise three');
end

% unpermute the factors

Z = Z(q,q);
D = D(q,q);
if nargout == 3
   W = W(q,q);
end

