/* sainvmex.c: implementation of SAINV preconditioner with MATLAB interface
* Usage:
*    [Z,D] = sainvmex(tol1,tol2,A);
*    [Z,D,W] = sainvmex(tol1,tol2,A,T);   % here T = A'
* Be warned that arguments are not stringently checked for errors, which
* may mean MATLAB will crash if you pass in erroneous arguments.
* Typically you would want tol1 smaller than tol2, and tol2 around 0.1.
* For example, tol1 = 1e-9, tol2 = 0.1.
*/

#include <math.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "mex.h"

static void SAINVsym (int n, void (*linop) (const void *, int, const int *,
     const double *, int *, int *, double *), const void *opdata,
     double droptol1, double droptol2, int **pZia, int **pZja, double **pZ,
     double **pD);
static void SAINVnon (int n, void (*linop) (const void *, int, const int *,
     const double *, int *, int *, double *), const void *opdata,
     void (*tranop) (const void *, int, const int *, const double *,
     int *, int *, double *), const void *trandata, double droptol1,
     double droptol2, int **pZia, int **pZja, double **pZ,
     int **pWia, int **pWja, double **pW, double **pD);
static void linop (const void *mat, int lenx, const int *nzx, const double *x,
   int *leny, int *nzy, double *y);
static void addsorted (double coeff, int lena, const int *nza, const double *a,
   int *leny, int *nzy, double *y, int n);

#define TOL1_IN   (prhs[0])
#define TOL2_IN   (prhs[1])
#define A_IN      (prhs[2])
#define T_IN      (prhs[3])
#define Z_OUT     (plhs[0])
#define D_OUT     (plhs[1])
#define W_OUT     (plhs[2])


void mexFunction (int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
   int i, n, *Zjc, *Zir, *Wjc, *Wir, *Djc, *Dir;
   double droptol1, droptol2, *Zpr, *Wpr, *Dpr;

   droptol1 = mxGetScalar (TOL1_IN);
   droptol2 = mxGetScalar (TOL2_IN);
   n = mxGetN(A_IN);

   if (nrhs == 3) {
      SAINVsym (n, linop, (void*) A_IN, droptol1, droptol2, &Zjc, &Zir, &Zpr,
         &Dpr);
   } else if (nrhs == 4) {
      SAINVnon (n, linop, (void*) A_IN, linop, (void*) T_IN, droptol1,
         droptol2, &Zjc, &Zir, &Zpr, &Wjc, &Wir, &Wpr, &Dpr);
      W_OUT = mxCreateSparse (n, n, 0, mxREAL);
      mxSetNzmax (W_OUT, Wjc[n]);
      mxSetJc (W_OUT, Wjc);
      mxSetIr (W_OUT, Wir);
      mxSetPr (W_OUT, Wpr);
   }

   Z_OUT = mxCreateSparse (n, n, 0, mxREAL);
   mxSetNzmax (Z_OUT, Zjc[n]);
   mxSetJc (Z_OUT, Zjc);
   mxSetIr (Z_OUT, Zir);
   mxSetPr (Z_OUT, Zpr);

   D_OUT = mxCreateSparse (n, n, 0, mxREAL);
   mxSetNzmax (D_OUT, n);
   Djc = (int*) mxCalloc (n+1, sizeof(int));
   Dir = (int*) mxCalloc (n, sizeof(int));
   for (i = 0; i < n; ++i) {
      Djc[i] = i;
      Dir[i] = i;
   }
   Djc[n] = n;
   mxSetJc (D_OUT, Djc);
   mxSetIr (D_OUT, Dir);
   mxSetPr (D_OUT, Dpr);
}


/* Sparse-sparse matvec */
static void linop (const void *mat, int lenx, const int *nzx, const double *x,
   int *leny, int *nzy, double *y)
{
   mxArray *A = (mxArray*) mat;
   int i, c, n, *jc, *ir;
   double *pr;

   n = mxGetN(A);
   jc = mxGetJc(A);
   ir = mxGetIr(A);
   pr = mxGetPr(A);

   *leny = 0;
   for (i = 0; i < lenx; ++i) {
      c = nzx[i];
      addsorted (x[i], jc[c+1]-jc[c], ir+jc[c], pr+jc[c], leny, nzy, y, n);
   }
}

/* Sparse axpy: y = coeff*a + y */
static void addsorted (double coeff, int lena, const int *nza, const double *a,
   int *leny, int *nzy, double *y, int n)
{
   int i, j, k, newleny;

   /* compute the new length of y */
   newleny = lena + *leny;
   i = 0; 
   j = 0;
   while (i < lena && j < *leny) {
      if (nza[i] < nzy[j]) {
         ++i;
      } else if (nza[i] > nzy[j]) {
         ++j;
      } else {  /* nza[i] == nzy[j]  which means overlap */
         --newleny;
         ++i; ++j;
      }
   }

   /* now merge into y, in reverse */
   k = newleny;
   i = lena-1;
   j = *leny-1;
   while (i >= 0 && j >= 0) {
      --k;
      if (nza[i] > nzy[j]) {
         y[k] = coeff*a[i];
         nzy[k] = nza[i];
         --i;
      } else if (nza[i] < nzy[j]) {
         y[k] = y[j];
         nzy[k] = nzy[j];
         --j;
      } else {  /* nza[i] == nzy[j]  which means overlap */
         y[k] = coeff*a[i] + y[j];
         nzy[k] = nza[i];
         --i; --j;
      }
   }
   while (i >= 0) {
      --k;
      y[k] = coeff*a[i];
      nzy[k] = nza[i];
      --i;
   }
   *leny = newleny;
}



/********************************* SAINV STUFF BELOW *************************/

#define INITNNZPERROW 10 /* how many nz per row to initially allocate */
#define ORDER_INDEP_UPDATE /* define this to make the updates independent */
                           /* of the ordering of previous columns; don't */
                           /* define this for SAINV-A which may be slightly */
                           /* higher quality. */

static void initSAINVfactor (int n, int *Zrowmaxlen, int *Zrowlen, int **Zrow,
   double **Zrowval, int *Zcolmaxlen, int *Zcollen, int **Zcol,
   double **Zcolval);
static void extractCol (int j, int *Zcollen, int **Zcol, double **Zcolval,
   int *Zrowlen, int **Zrow, double **Zrowval, double droptol);
static double sparseDot (int lenx, const int *nzx, const double *x,
   int leny, const int *nzy, const double *y);
static double infNorm (int lenx, const double *x);
static void sparseSparseMul (const int *collen, const int **col,
   const double **colval, int lenx, const int *nzx, const double *x,
   int *pleny, int *nzy, double *y);
static void insertForward (double coeff, int lenx, const int *nzx,
   const double *x, double droptol, int boundleny, int *pmaxleny, int *pleny,
   int **pnzy);
static void addReverseDrop (double coeff, int lenx, const int *nzx,
   const double *x, double droptol, int newleny, int *pmaxleny, int *pleny,
   int **pnzy, double **py);
static void updateSAINVfactor (int n, int j, int *Zrowmaxlen, int *Zrowlen,
   int **Zrow, double **Zrowval, int *Zcolmaxlen, int *Zcollen, int **Zcol,
   const double **Zcolval, int lenC, const int *nzC, const double *C,
   double droptol);
static void transferSAINVfactor (int n, const int *Zcollen, const int **Zcol,
   const double **Zcolval, int **pZia, int **pZja, double **pZ);
static void mmalloc (int numptrs, ...);
static void mfree (int numptrs, ...);


/* SAINVsym: do the (stabilized, outer-product) SAINV for symmetric matrices.
* n is the dimension of the problem.
* linop is the function for applying the linear operator, used in the form
*    linop (opdata, lenx, nzx, x, &leny, nzy, y) to apply the operator
*    (defined with data in opdata) to the sparse vector x giving result y,
*    another sparse vector. Both sparse vectors must be in sorted order.
* opdata is any data needed to specify the operator.
* droptol1 is used to sparsify outer-product updates
* droptol2 is used to filter the result.
* *pZia, *pZja, and *pZ store the upper triangular factor of the approximate
*    inverse column-by-column, not including the unit diagonal.
* D is a vector for the inverses of the pivots.
*/
static void SAINVsym (int n, void (*linop) (const void *, int, const int *,
     const double *, int *, int *, double *), const void *opdata,
     double droptol1, double droptol2, int **pZia, int **pZja, double **pZ,
     double **pD)
{
   /* The algorithm, in MATLAB notation but C indices, is:
   * Z = speye(n);
   * for j = 0:n-1
   *    Z(:,j) = sparsify2(Z(:,j));
   *    l = operator(Z(:,j));     % get column of approximate Cholesky factor
   *    D(j,j) = inv(l'*Z(:,j));  % compute inverse pivot
   *    L = l*D(j,j);             % rescale to approximate modified Cholesky 
   *    C = L'*Z(:,j+1:n-1);      % coefficients for update
   *    Z(:,j+1:n-1) = Z(:,j+1:n-1)-sparsify1(Z(:,j)*C); % outer-product update
   * end
   */

   /* for the dynamic part of Z (updates):
   * malloc each sparse row and sparse column separately. Store the
   * numerical values with the rows. The rows are stored in reverse sorted
   * order, the columns in forward sorted order.
   */
   int *Zrowmaxlen, *Zrowlen, **Zrow, *Zcolmaxlen, *Zcollen, **Zcol;
   double **Zrowval, **Zcolval;

   int i, j, lenl, *nzl, lenC, *nzC;
   double *D, *l, *C, pivot;

   /* default values */   
   *pZia = NULL;
   *pZja = NULL;
   *pZ = NULL;
   *pD = NULL;

   /* Do the initial allocations */
   mmalloc (13, &D, n*sizeof(double),       &Zrowmaxlen, n*sizeof(int),
                &Zrowlen, n*sizeof(int),    &Zrow, n*sizeof(int*),
                &Zcolmaxlen, n*sizeof(int), &Zcollen, n*sizeof(int),
                &Zcol, n*sizeof(int*),      &Zrowval, n*sizeof(double*),
                &Zcolval, n*sizeof(double*), 
                &nzl, n*sizeof(int),        &l, n*sizeof(double),
                &nzC, n*sizeof(int),        &C, n*sizeof(double));
   
   /* Allocate individual rows and columns for dynamic data structure, */
   /* putting the identity into Z */
   initSAINVfactor (n, Zrowmaxlen, Zrowlen, Zrow, Zrowval, Zcolmaxlen,
                    Zcollen, Zcol, Zcolval);
   
   /* main loop */
   for (j = 0; j < n; ++j) {
      /* extract j'th column of Z, filtered by droptol2 */
      extractCol (j, Zcollen, Zcol, Zcolval, Zrowlen, Zrow, Zrowval, droptol2);
      
      /* compute l = linop(Z_j) */
      linop (opdata, Zcollen[j], Zcol[j], Zcolval[j], &lenl, nzl, l);
      
      /* get inverse pivot D[j] = 1/(l,Z_j) */
      pivot = sparseDot (lenl, nzl, l, Zcollen[j], Zcol[j], Zcolval[j]);
      if (pivot == 0.0) {
         /* Oops - we broke down. Stick in a reasonable substitute */
         /* value for D[j] */ 
         D[j] = 1.0/infNorm (lenl, l);
         /* and then it doesn't make any sense to update the rest of Z */
         /* with this problematic column - so skip the update step. */
         /* A more robust version of the routine might report back this */
         /* breakdown, so that something can be done to the operator to */
         /* fix it... */
         continue;
      } else {
         D[j] = 1.0/pivot;
      }
      
      /* rescale L = l*D[j] */
      for (i = 0; i < lenl; ++i) {
         l[i] *= D[j];
      }
      
      /* find update coefficients C = L'*Z_(j+1:n-1) */
      sparseSparseMul (Zrowlen, Zrow, Zrowval, lenl, nzl, l, &lenC, nzC, C);
      
      /* Do outer-product update Z_(j+1:n) -= sparsify (Z_j*C, droptol1) */
      updateSAINVfactor (n, j, Zrowmaxlen, Zrowlen, Zrow, Zrowval,
             Zcolmaxlen, Zcollen, Zcol, Zcolval, lenC, nzC, C, droptol1);
   }
   
   /* Free up some of the stuff */
   for (i = 0; i < n; ++i) {
      mfree (2, Zrow[i], Zrowval[i]);
   }
   mfree (9, Zrowmaxlen, Zrowlen, Zrow, Zrowval, Zcolmaxlen, 
             nzl, l, nzC, C);

   /* Transfer the dynamic columns into the returned data structure */
   transferSAINVfactor (n, Zcollen, Zcol, Zcolval, pZia, pZja, pZ);
 
   *pD = D;

   /* free up rest */
   for (i = 0; i < n; ++i) {
      mfree (2, Zcol[i], Zcolval[i]);
   }
   mfree (3, Zcollen, Zcol, Zcolval);
}

/* SAINVnon: do the (stabilized?, outer-product) SAINV for nonsymmetric
* matrices.
* n is the dimension of the problem.
* linop is the function for applying the linear operator, used in the form
*    linop (opdata, lenx, nzx, x, &leny, nzy, y) to apply the operator
*    (defined with data in opdata) to the sparse vector x giving result y,
*    another sparse vector. Both sparse vectors must be in sorted order.
* opdata is any data needed to specify the operator.
* tranop and trandata are the same for the transposed operator.
* droptol1 is used to sparsify outer-product updates
* droptol2 is used to filter the result.
* *pZia, *pZja, and *pZ store the upper triangular factor of the approximate
*    inverse column-by-column, not including the unit diagonal.
* *pWia, *pWja, and *pW store the transpose of the lower triangular factor
*    column-by-column, not including the unit diagonal.
* D is a vector for the inverses of the pivots.
*/
static void SAINVnon (int n, void (*linop) (const void *, int, const int *,
     const double *, int *, int *, double *), const void *opdata,
     void (*tranop) (const void *, int, const int *, const double *,
     int *, int *, double *), const void *trandata, double droptol1,
     double droptol2, int **pZia, int **pZja, double **pZ,
     int **pWia, int **pWja, double **pW, double **pD)
{
   /* The algorithm, in MATLAB notation but C indices, is:
   * W = speye(n);
   * Z = speye(n);
   * for j = 0:n-1
   *    W(:,j) = sparsify2(W(:,j));
   *    Z(:,j) = sparsify2(Z(:,j));
   *    l = operator(Z(:,j));     % get column of approximate L factor
   *    u = transposeop(W(:,j));  % get column of approximate U' factor
   *    D(j,j) = inv(u'*Z(:,j));  % compute inverse pivot
   *    L = l*D(j,j);             % rescale to approximate modified LDU
   *    U = u*D(j,j);
   *    C = L'*W(:,j+1:n-1);      % coefficients for update
   *    W(:,j+1:n-1) = W(:,j+1:n-1)-sparsify1(W(:,j)*C); % outer-product update
   *    C = U'*Z(:,j+1:n-1);      % coefficients for update
   *    Z(:,j+1:n-1) = Z(:,j+1:n-1)-sparsify1(Z(:,j)*C); % outer-product update
   * end
   */

   /* for the dynamic part of Z and W (updates):
   * malloc each sparse row and sparse column separately. Store the
   * numerical values with the rows. The rows are stored in reverse sorted
   * order, the columns in forward sorted order.
   */
   int *Zrowmaxlen, *Zrowlen, **Zrow, *Zcolmaxlen, *Zcollen, **Zcol;
   double **Zrowval, **Zcolval;
   int *Wrowmaxlen, *Wrowlen, **Wrow, *Wcolmaxlen, *Wcollen, **Wcol;
   double **Wrowval, **Wcolval; 

   int i, j, lenl, *nzl, lenu, *nzu, lenC, *nzC;
   double *D, *l, *u, *C, pivot;

   /* default values */   
   *pZia = NULL;
   *pZja = NULL;
   *pZ = NULL;
   *pWia = NULL;
   *pWja = NULL;
   *pW = NULL;
   *pD = NULL;

   /* Do the initial allocations */
   mmalloc (23, &D, n*sizeof(double),       &Zrowmaxlen, n*sizeof(int),
                &Zrowlen, n*sizeof(int),    &Zrow, n*sizeof(int*),
                &Zcolmaxlen, n*sizeof(int), &Zcollen, n*sizeof(int),
                &Zcol, n*sizeof(int*),      &Zrowval, n*sizeof(double*),
                &Zcolval, n*sizeof(double*),&Wrowmaxlen, n*sizeof(int),
                &Wrowlen, n*sizeof(int),    &Wrow, n*sizeof(int*),
                &Wcolmaxlen, n*sizeof(int), &Wcollen, n*sizeof(int),
                &Wcol, n*sizeof(int*),      &Wrowval, n*sizeof(double*),
                &Wcolval, n*sizeof(double*),
                &nzl, n*sizeof(int),        &l, n*sizeof(double),
                &nzu, n*sizeof(int),        &u, n*sizeof(double),
                &nzC, n*sizeof(int),        &C, n*sizeof(double));
   
   /* Allocate individual rows and columns for dynamic data structure, */
   /* putting the identity into Z */
   initSAINVfactor (n, Zrowmaxlen, Zrowlen, Zrow, Zrowval, Zcolmaxlen,
                       Zcollen, Zcol, Zcolval);
   initSAINVfactor (n, Wrowmaxlen, Wrowlen, Wrow, Wrowval, Wcolmaxlen,
                       Wcollen, Wcol, Wcolval);
   
   /* main loop */
   for (j = 0; j < n; ++j) {
      /* extract j'th column of Z, filtered by droptol2 */
      extractCol (j, Zcollen, Zcol, Zcolval, Zrowlen, Zrow, Zrowval, droptol2);

      /* and the j'th column of W, also filtered by droptol2 */
      extractCol (j, Wcollen, Wcol, Wcolval, Wrowlen, Wrow, Wrowval, droptol2);
      
      /* compute l = linop(Z_j) */
      linop (opdata, Zcollen[j], Zcol[j], Zcolval[j], &lenl, nzl, l);

      /* compute u = tranop(W_j) */
      tranop (trandata, Wcollen[j], Wcol[j], Wcolval[j], &lenu, nzu, u);
      
      /* get inverse pivot D[j] = 1/(u,Z_j) = 1/(W_j,l) */
      if (lenu + Zcollen[j] < lenl + Wcollen[j]) {
         pivot = sparseDot (lenu, nzu, u, Zcollen[j], Zcol[j], Zcolval[j]);
      } else {
         pivot = sparseDot (Wcollen[j], Wcol[j], Wcolval[j], lenl, nzl, l);
      }
      if (pivot == 0.0) {
         /* Oops - we broke down. Stick in a reasonable substitute */
         /* value for D[j] */ 
         D[j] = 2.0 / (infNorm (lenl, l) + infNorm (lenu, u));
         /* and then it doesn't make any sense to update the rest of W, Z */
         /* with this problematic column - so skip the update step. */
         /* A more robust version of the routine might report back this */
         /* breakdown, so that something can be done to the operator to */
         /* fix it... */
         continue;
      } else {
         D[j] = 1.0/pivot;
      }
      
      /* rescale L = l*D[j], U = u*D[j] */
      for (i = 0; i < lenl; ++i) {
         l[i] *= D[j];
      }
      for (i = 0; i < lenu; ++i) {
         u[i] *= D[j];
      }
      
      /* find update coefficients C = L'*W_(j+1:n-1) */
      sparseSparseMul (Wrowlen, Wrow, Wrowval, lenl, nzl, l, &lenC, nzC, C);
      
      /* Do outer-product update W_(j+1:n) -= sparsify (W_j*C, droptol1) */
      updateSAINVfactor (n, j, Wrowmaxlen, Wrowlen, Wrow, Wrowval,
             Wcolmaxlen, Wcollen, Wcol, Wcolval, lenC, nzC, C, droptol1);
      
      /* find update coefficients C = U'*Z_(j+1:n-1) */
      sparseSparseMul (Zrowlen, Zrow, Zrowval, lenu, nzu, u, &lenC, nzC, C);
      
      /* Do outer-product update Z_(j+1:n) -= sparsify (Z_j*C, droptol1) */
      updateSAINVfactor (n, j, Zrowmaxlen, Zrowlen, Zrow, Zrowval,
             Zcolmaxlen, Zcollen, Zcol, Zcolval, lenC, nzC, C, droptol1);
   }
   
   /* Free up some of the stuff */
   for (i = 0; i < n; ++i) {
      mfree (4, Zrow[i], Zrowval[i], Wrow[i], Wrowval[i]);
   }
   mfree (16, Zrowmaxlen, Zrowlen, Zrow, Zrowval, Zcolmaxlen,
             Wrowmaxlen, Wrowlen, Wrow, Wrowval, Wcolmaxlen,
             nzl, l, nzu, u, nzC, C);

   /* Transfer the dynamic columns into the returned data structure */
   transferSAINVfactor (n, Wcollen, Wcol, Wcolval, pWia, pWja, pW);
   transferSAINVfactor (n, Zcollen, Zcol, Zcolval, pZia, pZja, pZ);
 
   *pD = D;

   /* free up rest */
   for (i = 0; i < n; ++i) {
      mfree (4, Zcol[i], Zcolval[i], Wcol[i], Wcolval[i]);
   }
   mfree (6, Zcollen, Zcol, Zcolval, Wcollen, Wcol, Wcolval);
}


/*********************** LOCAL SUPPORT FUNCTIONS *************************/


/* initSAINVfactor
* Allocate space for each row/column, and put the identity in there */
static void initSAINVfactor (int n, int *Zrowmaxlen, int *Zrowlen, int **Zrow,
   double **Zrowval, int *Zcolmaxlen, int *Zcollen, int **Zcol,
   double **Zcolval)
{
   int i, maxlen;
   
   for (i = 0; i < n; ++i) {
      maxlen = (i < INITNNZPERROW) ? i+1 : INITNNZPERROW;
      
      Zrowmaxlen[i] = maxlen;
      Zcolmaxlen[i] = maxlen;
      
      mmalloc (3, &(Zrow[i]), maxlen*sizeof(int),
                  &(Zrowval[i]), maxlen*sizeof(double),
                  &(Zcol[i]), maxlen*sizeof(int));
      
      /* put a diagonal 1 in the i'th row/col (except colval) */
      Zrowlen[i] = 1;
      Zrow[i][0] = i;
      Zrowval[i][0] = 1.0;
      
      Zcollen[i] = 1;
      Zcol[i][0] = i;
      Zcolval[i] = NULL;
   }
}

/* extractCol
* Remove the first (the j'th) column with filtering */
static void extractCol (int j, int *Zcollen, int **Zcol, double **Zcolval,
   int *Zrowlen, int **Zrow, double **Zrowval, double droptol)
{
   int i, nzc, r, rend;
   
   /* Allocate space for the numerical values in column j */
   Zcolval[j] = (double*) mxCalloc (Zcollen[j], sizeof(double));
   
   nzc = 0;
   for (i = 0; i < Zcollen[j]; ++i) {
      r = Zcol[j][i];
      rend = Zrowlen[r]-1;
      /* Check this entry against the drop tolerance */
      if (fabs(Zrowval[r][rend]) > droptol) {
         Zcolval[j][nzc] = Zrowval[r][rend];
         Zcol[j][nzc] = r;
         ++nzc;
      }
      /* Now remove the entry from row r */
      Zrowlen[r] = rend;
   }
   Zcollen[j] = nzc;
}

/* sparseDot
* Return the dot-product of two sorted sparse vectors */
static double sparseDot (int lenx, const int *nzx, const double *x,
   int leny, const int *nzy, const double *y)
{
   int i = 0, j = 0;
   double result = 0;
   
   while (i < lenx && j < leny) {
      if (nzx[i] < nzy[j]) {
         ++i;
      } else if (nzx[i] > nzy[j]) {
         ++j;
      } else {
         result += x[i]*y[j];
         ++i;
         ++j;
      }
   }
   return result;
}

/* infNorm
* Return the inf norm of the given sparse vector */
static double infNorm (int lenx, const double *x)
{
   int i;
   double maxentry = 0.0;
   
   for (i = 0; i < lenx; ++i) {
      if (x[i] > maxentry) {
         maxentry = x[i];
      }
   }
   return maxentry;
}

/* addReverse
* add a reverse-sorted sparse vector x (times coeff) to another one y,
* assuming y has adequate space already allocated. */
static void addReverse (double coeff, int lenx, const int *nzx,
   const double *x, int *pleny, int *nzy, double *y)
{
   int i, j, k, newleny;
   
   /* We'll do a linear time merge --- with sorted arrays holding the */
   /* sparse vectors, we probably can't do much better. If one of the */
   /* vectors is much longer than the other, we could take power-of-two */
   /* steps and backtrack instead of single steps, for some advantage. */
   
   /* do a count of common entries to find newleny */
   newleny = lenx + *pleny;
   i = lenx-1;
   j = *pleny-1;
   while (i >= 0 && j >= 0) {
      if (nzx[i] < nzy[j]) {
         --i;
      } else if (nzx[i] > nzy[j]) {
         --j;
      } else {
         --newleny;
         --i;
         --j;
      }
   }
   
   /* Now do the merge into y */
   k = newleny;
   i = lenx-1;
   j = *pleny-1;
   while (i >= 0 && j >= 0) {
      --k;
      if (nzx[i] < nzy[j]) {
         y[k] = coeff*x[i];
         nzy[k] = nzx[i];
         --i;
      } else if (nzx[i] > nzy[j]) {
         y[k] = y[j];
         nzy[k] = nzy[j];
         --j;
      } else {
         y[k] = coeff*x[i] + y[j];
         nzy[k] = nzy[j];
         --i;
         --j;
      }
   }
   /* If i >= 0 but j < 0, we add in the rest of x --- note we don't */
   /* have to do anything about the other case where i < 0 and j >= 0 */
   while (i >= 0) {
      --k;
      y[k] = coeff*x[i];
      nzy[k] = nzx[i];
      --i;
   }
   *pleny = newleny;
}

/* sparseSparseMul
* Do a fully sparse matvec y=A*x, with cols in reverse sorted order, x
* in forward sorted order, and the result y in reverse sorted order. We
* assume y has been allocated with adequate space already. */
static void sparseSparseMul (const int *Acollen, const int **Acol,
   const double **Acolval, int lenx, const int *nzx, const double *x,
   int *pleny, int *nzy, double *y)
{
   int i, c;
   
   *pleny = 0;
   for (i = 0; i < lenx; ++i) {
      c = nzx[i];
      addReverse (x[i], Acollen[c], Acol[c], Acolval[c],
         pleny, nzy, y);
   }
}

/* insertForward
* */
static void insertForward (double coeff, int lenx, const int *nzx,
   const double *x, double droptol, int boundleny, int *pmaxleny, int *pleny,
   int **pnzy)
{
   int i, j, k, newleny, newmaxleny, *newnzy;
   
   /* First determine how many nonzeros will end up in y */
   newleny = lenx + *pleny;
   i = 0;
   j = 0;
   while (i < lenx && j < *pleny) {
      if (nzx[i] < (*pnzy)[j]) {
         if (fabs(coeff*x[i]) <= droptol) {
            --newleny;   /* this new nonzero is too small to include */
         }
         ++i;
      } else if (nzx[i] > (*pnzy)[j]) {
         ++j;
      } else {
         --newleny;     /* nonzeros in common: only count it once */
         ++i;
         ++j;
      }
   }
   while (i < lenx) {
      if (fabs(coeff*x[i]) <= droptol) {
         --newleny;   /* this new nonzero is too small to include */
      }
      ++i;
   }    
   
   /* if we won't do anything later, can return early */
   if (newleny == *pleny) {
      return;
   }
   
   /* Check if we need to expand y */
   if (*pmaxleny < newleny) {
      newmaxleny = (3*newleny)/2;  /* keep some elbow room */
      if (newmaxleny > boundleny) {
         newmaxleny = boundleny;  /* but no sense in going above the bound */
      }
      /* Now allocate new array */
      newnzy = (int*) mxCalloc (newmaxleny, sizeof(int));
   } else {
      /* we didn't have to expand */
      newnzy = *pnzy;
   }
   
   /* do the merge, possibly in place */
   k = newleny;
   i = lenx-1;
   j = *pleny-1;
   while (i >= 0 && j >= 0) {
      if (nzx[i] < (*pnzy)[j]) {
         --k;
         newnzy[k] = (*pnzy)[j];
         --j;
      } else if (nzx[i] > (*pnzy)[j]) {
         if (fabs(coeff*x[i]) > droptol) {
            --k;
            newnzy[k] = nzx[i];
         }
         --i;
      } else {
         --k;
         newnzy[k] = nzx[i];
         --i;
         --j;
      }
   }
   while (i >= 0) {
      if (fabs(coeff*x[i]) > droptol) {
         --k;
         newnzy[k] = nzx[i];
      }
      --i;
   }
   if (newnzy != *pnzy) {   /* if we're not in place */
      /* finish the transfer */
      while (j >= 0) {
         --k;
         newnzy[k] = (*pnzy)[j];
         --j;
      }
      /* And update y */
      mxFree (*pnzy);
      *pnzy = newnzy;
      *pmaxleny = newmaxleny;
   }
   *pleny = newleny;
}

/* addReverseDrop
* */
static void addReverseDrop (double coeff, int lenx, const int *nzx,
   const double *x, double droptol, int boundleny, int *pmaxleny, int *pleny,
   int **pnzy, double **py)
{
   int i, j, k, newleny, newmaxleny, *newnzy;
   double *newy;
   
   /* First determine how many nonzeros will end up in y */
   newleny = lenx + *pleny;
   i = 0;
   j = 0;
   while (i < lenx && j < *pleny) {
      if (nzx[i] > (*pnzy)[j]) {
         if (fabs(coeff*x[i]) <= droptol) {
            --newleny;
         }
         ++i;
      } else if (nzx[i] < (*pnzy)[j]) {
         ++j;
      } else {
         --newleny;
         ++i;
         ++j;
      }
   }
   while (i < lenx) {
      if (fabs(coeff*x[i]) <= droptol) {
         --newleny;
      }
      ++i;
   }
   
   /* Check if we need to expand y */
   if (*pmaxleny < newleny) {
      newmaxleny = (3*newleny+1)/2;  /* keep some elbow room */
      if (newmaxleny > boundleny) {
         newmaxleny = boundleny;  /* but no sense in going above the bound */
      }
      /* Now allocate new array */
      mmalloc (2, &newnzy, newmaxleny*sizeof(int),
             &newy, newmaxleny*sizeof(double));
   } else {
      /* we didn't have to expand */
      newnzy = *pnzy;
      newy = *py;
   }
   
   /* do the merge and add, possibly in place */
   k = newleny;
   i = lenx-1;
   j = *pleny-1;
   while (i >= 0 && j >= 0) {
      if (nzx[i] > (*pnzy)[j]) {
         --k;
         newnzy[k] = (*pnzy)[j];
         newy[k] = (*py)[j];
         --j;
      } else if (nzx[i] < (*pnzy)[j]) {
         if (fabs(coeff*x[i]) > droptol) {
            --k;
            newnzy[k] = nzx[i];
            newy[k] = coeff*x[i];
         }
         --i;
      } else {
         --k;
         newnzy[k] = nzx[i];
#ifdef ORDER_INDEP_UPDATE
         if (fabs(coeff*x[i]) > droptol) {
            newy[k] = coeff*x[i] + (*py)[j];
         }
#else
         newy[k] = coeff*x[i] + (*py)[j];
#endif
         --i;
         --j;
      }
   }
   while (i >= 0) {
      if (fabs(coeff*x[i]) > droptol) {
         --k;
         newnzy[k] = nzx[i];
         newy[k] = coeff*x[i];
      }
      --i;
   }
   if (newnzy != *pnzy) {   /* if we're not in place */
      /* finish the transfer */
      while (j >= 0) {
         --k;
         newnzy[k] = (*pnzy)[j];
         newy[k] = (*py)[j];
         --j;
      }
      /* And update y */
      mxFree (*pnzy);
      *pnzy = newnzy;
      mxFree (*py);
      *py = newy;
      *pmaxleny = newmaxleny;
   }
   *pleny = newleny;
}

/* updateSAINVfactor
* Do a filtered outer-product update. Both Z rows and C are in reverse
* sorted order, but Z cols are in forward sorted order. */
static void updateSAINVfactor (int n, int j, int *Zrowmaxlen, int *Zrowlen,
   int **Zrow, double **Zrowval, int *Zcolmaxlen, int *Zcollen, int **Zcol,
   const double **Zcolval, int lenC, const int *nzC, const double *C,
   double droptol)
{
   int i, r, c;
   
   /* First add the update to the rows of Z */
   for (i = 0; i < Zcollen[j]; ++i) {
      r = Zcol[j][i];
      addReverseDrop (-Zcolval[j][i], lenC, nzC, C, droptol, n-r,
             &(Zrowmaxlen[r]), &(Zrowlen[r]), &(Zrow[r]), &(Zrowval[r]));
   }
   
   /* Now insert the new nonzeros in the columns of Z */
   for (i = 0; i < lenC; ++i) {
      c = nzC[i];
      insertForward (C[i], Zcollen[j], Zcol[j], Zcolval[j], droptol, c+1,
             &(Zcolmaxlen[c]), &(Zcollen[c]), &(Zcol[c]));
   }
}

/* transferSAINVfactor
* Move the factor from the dynamic data structure to a standard sparse
* format */
static void transferSAINVfactor (int n, const int *Zcollen, const int **Zcol,
   const double **Zcolval, int **pZia, int **pZja, double **pZ)
{
    int i, j, nnz;
    
    /* first count how many nonzeros we have */
    nnz = 0;
    for (i = 0; i < n; ++i) {
       nnz += Zcollen[i];
    }
    
    /* Allocate space */
    mmalloc (3, pZia, (n+1)*sizeof(int), pZja, nnz*sizeof(int),
             pZ, nnz*sizeof(double));
    
    /* Then transfer! */
    j = 0;
    (*pZia)[0] = 0;
    for (i = 0; i < n; ++i) {
       memcpy (*pZja + j, Zcol[i], Zcollen[i]*sizeof(int));
       memcpy (*pZ + j, Zcolval[i], Zcollen[i]*sizeof(double));
       j += Zcollen[i];
       (*pZia)[i+1] = j;
    }
}


/* multiple memory allocation */
static void mmalloc (int numptrs, ...)
{
   int i, j;
   va_list ap;
   void **p;
   size_t s;
   
   va_start (ap, numptrs);
   for (i = 0; i < numptrs; ++i) {
      p = va_arg (ap, void**);
      s = va_arg (ap, size_t);
      *p = mxCalloc (s, 1);
   }
   va_end (ap);
}

static void mfree (int numptrs, ...)
{
   int i;
   va_list ap;
   void *p;
   
   va_start (ap, numptrs);
   for (i = 0; i < numptrs; ++i) {
      p = va_arg (ap, void*);
      mxFree (p);
   }
   va_end (ap);
}

